package guru.kake.ronove

import guru.kake.xmlp.XMLParser
import java.io.FileInputStream
import javax.xml.stream.XMLEventReader
import javax.xml.stream.XMLInputFactory
import javax.xml.stream.events.XMLEvent


/**
 * Kanjidic2 Kanji dictionary character
 * @param literal representation as kanji symbol
 */
class Character(var literal: String = "") : Comparable<Character> {
    /** Readings and meanings */
    var rmgroups: MutableList<RMGroup> = mutableListOf()

    override fun equals(other: Any?): Boolean {
        if (other is Character) {
            return literal == other.literal
        }
        return super.equals(other)
    }

    override fun compareTo(other: Character): Int = literal.compareTo(other.literal)
    override fun hashCode(): Int = literal.hashCode()

}

/**
 * Combinations of readings and meanings for a dictionary entry
 */
class RMGroup {
    /** Readings of the character as Pair of reading type, reading */
    var reading: MutableList<Pair<String?, String>> = mutableListOf()
    /** Meadings of the character as Pair of reading language (null for English), meaning */
    var meaning: MutableList<Pair<String?, String>> = mutableListOf()
}

/**
 * JMdict phrase
 * @param japanese phrase in Japanese
 */
class Phrase(var japanese: String) : Comparable<Phrase> {
    /** Collection of meanings and translations for a phrase */
    var senses: MutableList<Sense> = mutableListOf()

    override fun equals(other: Any?): Boolean {
        if (other is Phrase) {
            return japanese == other.japanese
        }
        return super.equals(other)
    }
    override fun compareTo(other: Phrase): Int = japanese.compareTo(other.japanese)
    override fun hashCode(): Int = japanese.hashCode()
}

/**
 * Translation for a JMdict entry
 */
class Sense {
    /** List of translations */
    var glosses: MutableList<String> = mutableListOf()
}

/**
 * Collection of functions to parse several dictionary files provided by EDRDG.
 * Use functions under companion object, an instance of this class is not needed.
 * See https://www.edrdg.org/
 */
class DictionaryParser private constructor() {
    companion object {
        private val xmlInputFactory: XMLInputFactory = XMLInputFactory.newInstance()

        /**
         * Parses Kanjidic2 as provided by EDRDG
         * See https://www.edrdg.org/wiki/index.php/KANJIDIC_Project
         */
        fun parseKanjidic2(path: String): List<Character> {
            val parsed: MutableList<Character> = mutableListOf()
            var character: Character? = null
            var rmgroup: RMGroup? = null

            XMLParser.parse(path) {
                element("character") {
                    start { character = Character() }
                    end {
                        parsed.add(character!!)
                        character = null
                    }
                }
                element("literal") {
                    start { character!!.literal = content()!! }
                }
                element("reading_meaning") {
                    start { rmgroup = RMGroup() }
                    end {
                        character!!.rmgroups.add(rmgroup!!)
                        rmgroup = null
                    }
                }
                element("meaning") {
                    start {
                        rmgroup!!.meaning.add(Pair(
                            attribute("m_lang"),
                            content()!!
                        ))
                    }
                }
                element("reading") {
                    start {
                        rmgroup!!.reading.add(Pair(
                            attribute("r_type"),
                            content()!!
                        ))
                    }
                }
            }
            return parsed.toList()
        }

        /**
         * Parses JMdict as provided by EDRDG
         * See https://www.edrdg.org/wiki/index.php/JMdict-EDICT_Dictionary_Project
         */
        fun parseJMdict(path: String): List<Phrase> {
            //TODO: DSL
            val parsed: MutableList<Phrase> = mutableListOf()
            val phrases: MutableList<Phrase> = mutableListOf()
            var sense: Sense? = null

            val reader: XMLEventReader = xmlInputFactory.createXMLEventReader(FileInputStream(path))

            while (reader.hasNext()) {
                val nextEvent: XMLEvent = reader.nextEvent()
                when {
                    nextEvent.isStartElement -> {
                        val startElement = nextEvent.asStartElement()
                        when (startElement.name.localPart.toString()) {
                            //"entry" -> phrase = Character()
                            "keb" -> {
                                phrases.add(Phrase(reader.nextEvent().asCharacters().data))
                            }

                            "reb" -> {

                            }

                            "sense" -> {
                                sense = Sense()
                            }

                            "gloss" -> {
                                sense!!.glosses.add(reader.nextEvent().asCharacters().data)
                            }
                        }
                    }

                    nextEvent.isEndElement -> {
                        when (nextEvent.asEndElement().name.localPart.toString()) {
                            "entry" -> {
                                parsed.addAll(phrases)
                                phrases.clear()
                            }

                            "sense" -> {
                                phrases.forEach { it.senses.add(sense!!) }
                            }
                        }
                    }
                }
            }
            return parsed.toList()
        }
    }
}